// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"

#include "dco.hpp"

template<typename T>
T f(T p, T x);

/// least squares error of model predictions relative to observations
template<typename T>
T e(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  T e=0;
  int i=0;
  while (i<nobs) {
    e=e+pow(f(p,xobs(i))-yobs(i),2);
    i=i+1;
  } 
  return e;
}

/// first derivative of error wrt. model parameters by tangent AD
template<typename T>
T de_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // declarations and initializations
  using TT=dco::gt1s<T>::type;
  TT pTT=p, eTT;
  // seeding tangent of model parameter
  dco::derivative(pTT)=1;
  // computing tangent of error
  eTT=e<TT>(pTT,xobs,yobs);
  // harvesting tangent of model parameter
  return dco::derivative(eTT);
}

/// derivative of derivative of error wrt. model parameters wrt. model parameters by finite differences
template<typename T>
T dde_dp_dp_cfd(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // size of perturbation
  T h=1e-7;
  if (p!=0) {
    h=h*fabs(p);
  }
  // central finite difference of first derivative
  return (de_dp(p+h,xobs,yobs)-de_dp(p-h,xobs,yobs))/(h+h);
}

/// derivative of derivative of error wrt. model parameters wrt. model parameters by tangent AD
template<typename T>
T dde_dp_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // declarations and initializations
  using TT=dco::gt1s<T>::type;
  TT pTT=p, deTT;
  // seeding tangent model parameter
  dco::derivative(pTT)=1;
  // computing tangent of first derivative of error wrt. model parameters
  deTT=de_dp<TT>(pTT,xobs,yobs);
  // harvesting tangent of first derivative of error wrt. model parameters
  return dco::derivative(deTT);
}

/// first derivative of error wrt. observations by adjoint AD
template<typename T>
VT<T> de_dobs(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // declarations and initializations
  VT<T> dedobs(2*nobs); // result
  using AT=dco::ga1s<T>::type;
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  AT eAT;
  VT<AT> xobsAT(nobs), yobsAT(nobs);
  int i=0,j;
  while (i<nobs) {
    dco::value(xobsAT(i))=xobs(i); 
    dco::value(yobsAT(i))=yobs(i);
    i=i+1;
  }
  // registration of observations with tape
  i=0;
  while (i<nobs) {
    tape->register_variable(xobsAT(i));
    tape->register_variable(yobsAT(i));
    i=i+1;
  }
  // recording of tape
  eAT=e<AT>(p,xobsAT,yobsAT);
  tape->register_output_variable(eAT);
  // seeding adjoint error
  dco::derivative(eAT)=1;
  // interpretation of tape
  tape->interpret_adjoint();
  // harvesting adjoint observations
  j=0;
  i=0;
  while (i<nobs) {
    dedobs(j)=dco::derivative(xobsAT(i));
    j=j+1;
    dedobs(j)=dco::derivative(yobsAT(i));
    j=j+1; 
    i=i+1;
  }
  return dedobs;
}

/// derivative of derivative of error wrt. observations wrt. model parameters by tangent AD
template<typename T>
VT<T> dde_dobs_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // declarations and initializations
  VT<T> ddedobsdp(2*nobs); // result
  using TT=dco::gt1s<T>::type;
  TT pTT=p;
  VT<TT> deTT(nobs);
  // seeding tangent of model parameter
  dco::derivative(pTT)=1;
  // computing tangent of first derivative of error wrt. observations
  deTT=de_dobs<TT>(pTT,xobs,yobs);
  // harvesting tangent of first derivative of error wrt. observations
  int i=0;
  while (i<2*nobs) {
    ddedobsdp(i)=dco::derivative(deTT(i));
    i=i+1;
  }
  return ddedobsdp;
}
