// naumann@stce.rwth-aachen.de

#include <Eigen/Dense>

template<typename T>
using VT=Eigen::VectorX<T>;

#include "dco.hpp"

#include <cmath>
using namespace std;

int nobs=5; // number of observations
double eps=1e-7; // base accuracy of iterative methods

template<typename T>
T f(T p, T x) {
  return sin(p*pow(x,2));
}

template<typename T>
T e(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  T e=0;
  int i=0;
  while (i<nobs) {
    e=e+pow(f(p,xobs(i))-yobs(i),2);
    i=i+1;
  } 
  return e;
}

template<typename T>
T de_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // declarations and initializations
  using TT=dco::gt1s<T>::type;
  TT pTT=p, eTT;
  // seeding tangent of model parameter
  dco::derivative(pTT)=1;
  // computing tangent of error
  eTT=e<TT>(pTT,xobs,yobs);
  // harvesting tangent of model parameter
  return dco::derivative(eTT);
}

template<typename T>
T dde_dp_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // declarations and initializations
  using TT=dco::gt1s<T>::type;
  TT pTT=p, deTT;
  // seeding tangent model parameter
  dco::derivative(pTT)=1;
  // computing tangent of first derivative of error wrt. model parameters
  deTT=de_dp<TT>(pTT,xobs,yobs);
  // harvesting tangent of first derivative of error wrt. model parameters
  return dco::derivative(deTT);
}

template<typename T>
VT<T> r(T p, VT<T> xobs, VT<T> yobs) {
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  VT<T> r_(nobs);
  int i=0;
  while (i<nobs) {
    r_(i)=f(p,xobs(i))-yobs(i);
    i=i+1;
  }
  return r_;
}

template<typename T> 
VT<T> dr_dp(T p, VT<T> xobs) { 
  assert(xobs.size()==nobs);
  VT<T> drdp(nobs);
  using TT=dco::gt1s<T>::type;
  TT pTT, yTT;
  int i=0;
  while (i<nobs) {
    pTT=p;
    dco::derivative(pTT)=1;
    yTT=f<TT>(pTT,xobs(i));
    drdp(i)=dco::derivative(yTT);
    i=i+1;
  }
  return drdp;
}

template<typename T>
T linear_regression(VT<T> drdp, VT<T> yobs) {
  assert(nobs==drdp.size()&&drdp.size()==yobs.size());
  VT<T> v=drdp;
  v(0)=v(0)-drdp(0)/fabs(drdp(0))*drdp.norm();
  VT<T> n=yobs-2*v/(v.dot(v))*(v.dot(yobs));
  return n(0)/drdp.norm();
}

template<typename T>
T fit(T p, VT<T> xobs, VT<T> yobs) {
  assert(nobs==xobs.size()&&xobs.size()==yobs.size());
  T delta_p=1;
  while (fabs(delta_p)>eps) {
    delta_p=linear_regression(dr_dp(p,xobs),r(p,xobs,yobs));
    p=p-delta_p;
  }
  return p;
}

template<typename T>
void plot(VT<T> x, VT<T> y) {
  assert(x.size()==y.size());
  ofstream o("data");
  for (int i=0;i<x.size();++i) {
    o << x(i) << ' ' << y(i) << endl;
  }
}

int main() {
  using T=double;
  T p=1;
  VT<T> x=VT<T>::Random(nobs), y=VT<T>::Random(nobs);
  plot(x,y);
  p=fit(p,x,y);
  cout << "p=" << p << endl;
  cout << "de_dp=" << de_dp(p,x,y) << endl;
  cout << "dde_dp_dp=" << dde_dp_dp(p,x,y) << endl;
  return 0;
}
