// naumann@stce.rwth-aachen.de

#include <Eigen/Dense>

template<typename T>
using VT=Eigen::VectorX<T>;

#include "dco.hpp"

#include <cmath>
using namespace std;

int nobs=5; // number of observations
double eps=1e-7; // base accuracy of iterative methods

template<typename T>
T f(T p, T x) {
  return sin(p*pow(x,2));
}

template<typename T>
T e(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  T e=0;
  int i=0;
  while (i<nobs) {
    e=e+pow(f(p,xobs(i))-yobs(i),2);
    i=i+1;
  } 
  return e;
}

template<typename T>
T de_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // declarations and initializations
  using TT=dco::gt1s<T>::type;
  TT pTT=p, eTT;
  // seeding tangent of model parameter
  dco::derivative(pTT)=1;
  // computing tangent of error
  eTT=e<TT>(pTT,xobs,yobs);
  // harvesting tangent of model parameter
  return dco::derivative(eTT);
}

template<typename T>
T dde_dp_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // declarations and initializations
  using TT=dco::gt1s<T>::type;
  TT pTT=p, deTT;
  // seeding tangent model parameter
  dco::derivative(pTT)=1;
  // computing tangent of first derivative of error wrt. model parameters
  deTT=de_dp<TT>(pTT,xobs,yobs);
  // harvesting tangent of first derivative of error wrt. model parameters
  return dco::derivative(deTT);
}

/// derivative of error wrt. observations by adjoint AD
template<typename T>
VT<T> de_dobs(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // adjoint type
  using AT=dco::ga1s<T>::type;
  // smart pointer to tape
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  // activation
  AT e_a;
  VT<AT> xobs_a(nobs), yobs_a(nobs);
  int i=0;
  while (i<nobs) {
    xobs_a(i)=xobs(i); 
    yobs_a(i)=yobs(i);
    i=i+1;
  }
  // registration of observations with tape
  i=0;
  while (i<nobs) {
    tape->register_variable(xobs_a(i));
    tape->register_variable(yobs_a(i));
    i=i+1;
  }
  // overloaded computation of error records tape
  e_a=e<AT>(p,xobs_a,yobs_a);
  // seeding derivative of error
  dco::derivative(e_a)=1;
  // interpretation of tape
  tape->interpret_adjoint();
  // harvesting derivative of error wrt. obserations
  VT<T> dedobs(2*nobs); 
  int j=0;
  i=0;
  while (i<nobs) {
    dedobs(j)=dco::derivative(xobs_a(i)); j=j+1;
    dedobs(j)=dco::derivative(yobs_a(i)); j=j+1; 
    i=i+1;
  }
  return dedobs;
}

/// derivative of derivative of error wrt. observations wrt. model parameter by tangent AD
template<typename T>
VT<T> dde_dobs_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // tangent type
  using TT=dco::gt1s<T>::type;
  // activation
  TT p_t=p;
  VT<TT> dedobs_t(nobs);
  // seeding derivative of model parameter
  dco::derivative(p_t)=1;
  // overloaded compution of derivative of error wrt. observations
  dedobs_t=de_dobs<TT>(p_t,xobs,yobs);
  // harvesting derivative of derivative of error wrt. observations wrt. model parameter
  VT<T> ddedobsdp(2*nobs); 
  int i=0;
  while (i<2*nobs) {
    ddedobsdp(i)=dco::derivative(dedobs_t(i));
    i=i+1;
  }
  return ddedobsdp;
}

template<typename T>
VT<T> r(T p, VT<T> xobs, VT<T> yobs) {
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  VT<T> r_(nobs);
  int i=0;
  while (i<nobs) {
    r_(i)=f(p,xobs(i))-yobs(i);
    i=i+1;
  }
  return r_;
}

template<typename T> 
VT<T> dr_dp(T p, VT<T> xobs) { 
  assert(xobs.size()==nobs);
  VT<T> drdp(nobs);
  using TT=dco::gt1s<T>::type;
  TT pTT, yTT;
  int i=0;
  while (i<nobs) {
    pTT=p;
    dco::derivative(pTT)=1;
    yTT=f<TT>(pTT,xobs(i));
    drdp(i)=dco::derivative(yTT);
    i=i+1;
  }
  return drdp;
}

template<typename T>
T linear_regression(VT<T> drdp, VT<T> yobs) {
  assert(drdp.size()==yobs.size());
  return drdp.dot(yobs)/drdp.dot(drdp);
}

template<typename T>
T fit(T p, VT<T> xobs, VT<T> yobs) {
  assert(nobs==xobs.size()&&xobs.size()==yobs.size());
  T delta_p=1;
  while (fabs(delta_p)>eps) {
    delta_p=linear_regression(dr_dp(p,xobs),r(p,xobs,yobs));
    p=p-delta_p;
  }
  return p;
}

template<typename T>
void plot(VT<T> x, VT<T> y) {
  assert(x.size()==y.size());
  ofstream o("data");
  for (int i=0;i<x.size();++i) {
    o << x(i) << ' ' << y(i) << endl;
  }
}

template<typename T>
T fit(T p, VT<T> xobs, VT<T> yobs);

/// stability analysis by symbolic differentiation
template<typename T>
VT<T> dp_dobs_sym(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // result of differentiation of first-order optimality condition
  return -dde_dobs_dp(p,xobs,yobs)/dde_dp_dp(p,xobs,yobs);
}

/// stability analysis by central finite differences
template<typename T>
VT<T> dp_dobs_cfd(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // return value
  VT<T> dpdobs(2*nobs);
  int i=0, j=0;
  while (i<nobs) {
    T xobs_i_save=xobs(i), yobs_i_save=yobs(i), pp, pm; 
    // perturbing approx. midpoint of mantissa
    T h=eps;
    if (xobs(i)!=0) {
      h=h*fabs(xobs(i));
    }
    // central finite difference quotient
    xobs(i)=xobs(i)+h;
    pp=fit(p,xobs,yobs);
    xobs(i)=xobs_i_save-h;
    pm=fit(p,xobs,yobs);
    xobs(i)=xobs_i_save;
    dpdobs(j)=(pp-pm)/(2*h); j=j+1;
    // perturbing approx. midpoint of mantissa
    h=eps;
    if (yobs(i)!=0) {
      h=h*fabs(yobs(i));
    }
    // central finite difference quotient
    yobs(i)=yobs(i)+h;
    pp=fit(p,xobs,yobs);
    yobs(i)=yobs_i_save-h;
    pm=fit(p,xobs,yobs);
    yobs(i)=yobs_i_save;
    dpdobs(j)=(pp-pm)/(2*h); j=j+1;
    i=i+1;
  }
  return dpdobs;
}

/// stability analysis by tangent AD
template<typename T>
VT<T> dp_dobs_tad(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // tangent type
  using TT=dco::gt1s<T>::type;
  // activation
  TT p_t;
  VT<TT> xobs_t(nobs), yobs_t(nobs);
  int i=0;
  while (i<nobs) {
    xobs_t(i)=xobs(i); 
    yobs_t(i)=yobs(i);
    i=i+1;
  }
  // return value
  VT<T> dpdobs(2*nobs);
  int j=0;
  i=0;
  while (i<nobs) {
    // tangent in unit direction of current point of observation
    p_t=p; // reset model parameter to its initial value
    // seed
    dco::derivative(xobs_t(i))=1;
    // overloaded parameter fitting
    p_t=fit(p_t,xobs_t,yobs_t); // all arguments of same type
    // harvest
    dpdobs(j)=dco::derivative(p_t); j=j+1;
    // enable correct next seed
    dco::derivative(xobs_t(i))=0; 
    // tangent in unit direction of current observed value
    p_t=p; // reset model parameter to its initial value
    // seed
    dco::derivative(yobs_t(i))=1;
    // overloaded parameter fitting
    p_t=fit(p_t,xobs_t,yobs_t);
    // harvest
    dpdobs(j)=dco::derivative(p_t); j=j+1;
    // enable correct next seed
    dco::derivative(yobs_t(i))=0;
    i=i+1;
  }
  return dpdobs;
}

/// stability analysis by adjoint AD
template<typename T>
VT<T> dp_dobs_aad(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // adjoint type
  using AT=dco::ga1s<T>::type;
  // tape
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  // activation and registration
  AT p_a=p;
  VT<AT> xobs_a(nobs), yobs_a(nobs);
  int i=0;
  while (i<nobs) {
    xobs_a(i)=xobs(i); 
    tape->register_variable(xobs_a(i));
    yobs_a(i)=yobs(i);
    tape->register_variable(yobs_a(i));
    i=i+1;
  }
  // overloaded parameter fitting
  p_a=fit(p_a,xobs_a,yobs_a);
  // seed
  dco::derivative(p_a)=1;
  // interpretation of tape
  tape->interpret_adjoint();
  // harvest
  VT<T> dpdobs(2*nobs);
  int j=0;
  i=0;
  while (i<nobs) {
    dpdobs(j)=dco::derivative(xobs_a(i)); j=j+1; 
    dpdobs(j)=dco::derivative(yobs_a(i)); j=j+1; 
    i=i+1;
  }
  return dpdobs;
}

int main() {
  using T=double;
  T p=1;
  VT<T> x=VT<T>::Random(nobs), y=VT<T>::Random(nobs);
  p=fit(p,x,y);
  cout << "p=" << p << endl;
  cout << "de_dp=" << de_dp(p,x,y) << endl;
  cout << "dde_dp_dp=" << dde_dp_dp(p,x,y) << endl;
  cout << "dpdobs (TAD)=" << dp_dobs_tad(p,x,y).transpose() << endl;
  return 0;
}
